/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _DOMAIN_ACCESS_CONTROL_WIIU_H_
#define _DOMAIN_ACCESS_CONTROL_WIIU_H_

#include <list>

#include "config.h"
#include "WTFString.h"

namespace WebCore
{
namespace DomainAccessControl 
{

enum
{
	DOMAIN_ACCESS_NOT_ENCRYPTED = -3,
	DOMAIN_ACCESS_NOT_WHITELISTED = -2,
	DOMAIN_ACCESS_NOT_ACQUIRED_AUTH_TOKEN = -1,
	DOMAIN_ACCESS_OK = 0
};

typedef struct DomainAccess
{
	String origin;
	bool subdomains;
} DomainAccess;

std::list<DomainAccess>& domainAccessList(void);

void AllowUnrestrictedDomains();
void ClearDomainWhitelist();
bool AddDomainToWhitelist(const String &origin, bool subdomains);
bool AddDomainToWhitelist(const DomainAccess &domain);
bool IsDomainAccessible(const String &origin, int *pError = NULL);

}
}

#endif