/*
 *  Copyright (C) 2007 Alp Toker <alp@atoker.com>
 *  Copyright (C) 2008 Nuanti Ltd.
 *  Copyright (C) 2008 INdT - Instituto Nokia de Tecnologia
 *  Copyright (C) 2009-2010 ProFUSION embedded systems
 *  Copyright (C) 2009-2010 Samsung Electronics
 *  Copyright (c) 2012 NINTENDO.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "EditorClientWiiU.h"
#include "EditCommand.h"
#include "Editor.h"
#include "EventNames.h"
#include "FocusController.h"
#include "Frame.h"
#include "KeyboardEvent.h"
#include "NotImplemented.h"
#include "Page.h"
#include "Settings.h"
#include "PlatformKeyboardEvent.h"
#include "WebViewWiiU.h"
#include "WindowsKeyboardCodes.h"

#include <PlatformInput.h>
#include <PlatformSoftwareKeyboard.h>
using namespace LibPlatform;

#include <PlatformOS.h>
#include <PlatformDefines.h>
using namespace LibPlatform;

using namespace WebCore;

namespace WebCore {

void EditorClientWiiU::willSetInputMethodState()
{
	notImplemented();
}

void EditorClientWiiU::setInputMethodState(bool active)
{
	if (active)
		m_view->softwareKeyboardInvoke(this);
	else
		m_view->softwareKeyboardDismiss();
}

bool EditorClientWiiU::shouldDeleteRange(Range*)
{
    notImplemented();
    return true;
}

bool EditorClientWiiU::shouldShowDeleteInterface(HTMLElement*)
{
    return false;
}

bool EditorClientWiiU::isContinuousSpellCheckingEnabled()
{
    notImplemented();
    return false;
}

bool EditorClientWiiU::isGrammarCheckingEnabled()
{
    notImplemented();
    return false;
}

int EditorClientWiiU::spellCheckerDocumentTag()
{
    notImplemented();
    return 0;
}

bool EditorClientWiiU::shouldBeginEditing(Range*)
{
	notImplemented();
    return true;
}

bool EditorClientWiiU::shouldEndEditing(Range*)
{
	notImplemented();
    return true;
}

bool EditorClientWiiU::shouldInsertText(const String& str, Range *pRange, EditorInsertAction action)
{
	notImplemented();
    return true;
}

// this is just for debugging
static void printRange(Range *pRange)
{
	OS::Report("Range (%p)\n", pRange);
	if (pRange)
	{
		Position startPosition = pRange->startPosition();
		Position endPosition = pRange->endPosition();
		OS::Report("  startPosition: %d\n", startPosition.getOffset());
		OS::Report("  endPosition: %d\n", endPosition.getOffset());
		/*ExceptionCode exc = 0;

		Document *pOwnerDocument = pRange->ownerDocument();
		Node *pStartContainer = pRange->startContainer();
		Node *pEndContainer = pRange->endContainer();
		Node *pFirstNode = pRange->firstNode();
		Node *pPastLastNode = pRange->pastLastNode();
		Node *pShadowTreeRootNode = pRange->shadowTreeRootNode();
		Position startPosition = pRange->startPosition();
		Position endPosition = pRange->endPosition();

		OS::Report("  ownerDocument: %p\n", pOwnerDocument);
		if (pOwnerDocument)
		{
			// Element *getElementById
			OS::Report("    doctype: %p\n", pOwnerDocument->doctype());
			OS::Report("    documentElement: %p\n", pOwnerDocument->documentElement());
			OS::Report("    documentURI: %s\n", pOwnerDocument->documentURI().latin1().data());
		}
		OS::Report("  startContainer: %p\n", pStartContainer);
		OS::Report("  endContainer: %p\n", pEndContainer);
		if (pStartContainer && pEndContainer)
			OS::Report("  commonAncestorContainer: %p\n", pRange->commonAncestorContainer(pStartContainer, pEndContainer));
		OS::Report("  firstNode: %p\n", pFirstNode);
		OS::Report("  pastLastNode: %p\n", pPastLastNode);
		OS::Report("  shadowTreeRootNode: %p\n", pShadowTreeRootNode);
		OS::Report("  toString: %s\n", pRange->toString(exc).latin1().data());
		OS::Report("  toHTML: %s\n", pRange->toHTML().latin1().data());
		OS::Report("  text: %s\n", pRange->text().latin1().data());
		OS::Report("  startPosition offset: %d\n", startPosition.getOffset());
		OS::Report("  endPosition offset: %d\n", endPosition.getOffset());*/
	}
}

bool EditorClientWiiU::shouldChangeSelectedRange(Range *pFromRange, Range *pToRange, EAffinity affinity, bool bStillSelecting)
{
	//#8788 pToRange could be 0 after the toNormalizedRange() function when selectionType == NoSelection 
	if (pToRange!=0)
	{
		Position startPosition = pToRange->startPosition();
		Position endPosition = pToRange->endPosition();

		SoftwareKeyboard::GetInstance()->SetSelection(startPosition.getOffset(), endPosition.getOffset());
	}

	/*OS::Report("--- printing pFromRange ---\n");
	printRange(pFromRange);
	OS::Report("--- printing pToRange ---\n");
	printRange(pToRange);*/

    notImplemented();
	return true;
}

bool EditorClientWiiU::shouldApplyStyle(CSSStyleDeclaration*, Range*)
{
    notImplemented();
    return true;
}

bool EditorClientWiiU::shouldMoveRangeAfterDelete(Range*, Range*)
{
    notImplemented();
    return true;
}

void EditorClientWiiU::didBeginEditing()
{
	notImplemented();
}

void EditorClientWiiU::respondToChangedContents()
{
    notImplemented();
}

void EditorClientWiiU::respondToChangedSelection(Frame* coreFrame)
{
	m_focusedFrame = coreFrame;
}

void EditorClientWiiU::didEndEditing()
{
	notImplemented();
}

void EditorClientWiiU::didWriteSelectionToPasteboard()
{
    notImplemented();
}

void EditorClientWiiU::didSetSelectionTypesForPasteboard()
{
    notImplemented();
}

void EditorClientWiiU::registerCommandForUndo(WTF::PassRefPtr<EditCommand>)
{
    notImplemented();
}

void EditorClientWiiU::registerCommandForRedo(WTF::PassRefPtr<EditCommand>)
{
    notImplemented();
}

void EditorClientWiiU::clearUndoRedoOperations()
{
    notImplemented();
}

bool EditorClientWiiU::canPaste(Frame*, bool defaultValue) const
{
    return defaultValue;
}

bool EditorClientWiiU::canCopyCut(Frame*, bool defaultValue) const
{
    return defaultValue;
}

bool EditorClientWiiU::canUndo() const
{
    notImplemented();
    return false;
}

bool EditorClientWiiU::canRedo() const
{
    notImplemented();
    return false;
}

void EditorClientWiiU::undo()
{
    notImplemented();
}

void EditorClientWiiU::redo()
{
    notImplemented();
}

bool EditorClientWiiU::shouldInsertNode(Node*, Range*, EditorInsertAction)
{
    notImplemented();
    return true;
}

void EditorClientWiiU::pageDestroyed()
{
    delete this;
}

bool EditorClientWiiU::smartInsertDeleteEnabled()
{
    notImplemented();
    return false;
}

bool EditorClientWiiU::isSelectTrailingWhitespaceEnabled()
{
    notImplemented();
    return false;
}

void EditorClientWiiU::toggleContinuousSpellChecking()
{
    notImplemented();
}

void EditorClientWiiU::toggleGrammarChecking()
{
    notImplemented();
}

bool EditorClientWiiU::handleEditingTextInput(const String& text)
{
	Frame* frame = m_focusedFrame;
	if (!frame) {
		return false;
	}

	bool result = frame->editor()->setSelectionOffsets(0, 100);
	//frame->editor()->replaceSelectionWithText(text, false, true);
	return true;
}

String EditorClientWiiU::getEditingText()
{
	Frame* frame = m_focusedFrame;
	if (!frame) {
		return String();
	}

	int selectionStart = 0;
	int selectionEnd = 0;

	frame->editor()->getSelectionOffsets(&selectionStart, &selectionEnd);
	frame->editor()->setSelectionOffsets(0, INT_MAX);
	String text = frame->editor()->selectedText();
	frame->editor()->setSelectionOffsets(selectionStart, selectionEnd);

	return text;
}

void EditorClientWiiU::setCursorPos(unsigned int cursorPos)
{
	Frame* frame = m_focusedFrame;
	if (!frame)
		return;

	frame->editor()->setSelectionOffsets(cursorPos, cursorPos);
}

static const unsigned CtrlKey = 1 << 0;
static const unsigned AltKey = 1 << 1;
static const unsigned ShiftKey = 1 << 2;

struct KeyDownEntry {
    unsigned virtualKey;
    unsigned modifiers;
    const char* name;
};

struct KeyPressEntry {
    unsigned charCode;
    unsigned modifiers;
    const char* name;
};

static const KeyDownEntry keyDownEntries[] = {
    {VK_LEFT,		0,                  "MoveLeft"                                    },
    {VK_LEFT,		ShiftKey,           "MoveLeftAndModifySelection"                  },
    {VK_LEFT,		CtrlKey,            "MoveWordLeft"                                },
    {VK_LEFT,		CtrlKey | ShiftKey, "MoveWordLeftAndModifySelection"              },
    {VK_RIGHT,		0,                  "MoveRight"                                   },
    {VK_RIGHT,		ShiftKey,           "MoveRightAndModifySelection"                 },
    {VK_RIGHT,		CtrlKey,            "MoveWordRight"                               },
    {VK_RIGHT,		CtrlKey | ShiftKey, "MoveWordRightAndModifySelection"             },
    {VK_UP,			0,                  "MoveUp"                                      },
    {VK_UP,			ShiftKey,           "MoveUpAndModifySelection"                    },
	{VK_PRIOR,		ShiftKey,           "MovePageUpAndModifySelection"                },
    {VK_DOWN,		0,                  "MoveDown"                                    },
    {VK_DOWN,		ShiftKey,           "MoveDownAndModifySelection"                  },
    {VK_NEXT,		ShiftKey,           "MovePageDownAndModifySelection"              },
    {VK_PRIOR,		0,                  "MovePageUp"                                  },
    {VK_NEXT,		0,                  "MovePageDown"                                },
    {VK_HOME,		0,                  "MoveToBeginningOfLine"                       },
    {VK_HOME,		ShiftKey,           "MoveToBeginningOfLineAndModifySelection"     },
    {VK_HOME,		CtrlKey,            "MoveToBeginningOfDocument"                   },
    {VK_HOME,		CtrlKey | ShiftKey, "MoveToBeginningOfDocumentAndModifySelection" },
    {VK_END,		0,                  "MoveToEndOfLine"                             },
    {VK_END,		ShiftKey,           "MoveToEndOfLineAndModifySelection"           },
    {VK_END,		CtrlKey,            "MoveToEndOfDocument"                         },
    {VK_END,		CtrlKey | ShiftKey, "MoveToEndOfDocumentAndModifySelection"       },
    {VK_BACK,		0,                  "DeleteBackward"                              },
    {VK_BACK,		ShiftKey,           "DeleteBackward"                              },
    {VK_DELETE,		0,                  "DeleteForward"                               },
    {VK_BACK,		CtrlKey,            "DeleteWordBackward"                          },
    {VK_DELETE,		CtrlKey,            "DeleteWordForward"                           },
    {'B',			CtrlKey,            "ToggleBold"                                  },
    {'I',			CtrlKey,            "ToggleItalic"                                },
    {VK_ESCAPE,		0,                  "Cancel"                                      },
    {VK_OEM_PERIOD,	CtrlKey,			"Cancel"                                      },
    {VK_TAB,		0,                  "InsertTab"                                   },
    {VK_TAB,		ShiftKey,           "InsertBacktab"                               },
    {VK_RETURN,		0,                  "InsertNewline"                               },
    {VK_RETURN,		CtrlKey,            "InsertNewline"                               },
    {VK_RETURN,		AltKey,             "InsertNewline"                               },
    {VK_RETURN,		AltKey | ShiftKey,  "InsertNewline"},
};

static const KeyPressEntry keyPressEntries[] = {
    {Keyboard_Tab,					0,                  "InsertTab"							},
    {Keyboard_Tab,					ShiftKey,           "InsertBacktab"						},
    {Keyboard_Enter,				0,                  "InsertNewline"						},
    {Keyboard_Enter,				CtrlKey,            "InsertNewline"						},
    {Keyboard_Enter,				AltKey,             "InsertNewline"						},
    {Keyboard_Enter,				AltKey | ShiftKey,  "InsertNewline"						},
	{Keyboard_Backspace,			0,					"DeleteBackward"					},
    {Keyboard_Backspace,			ShiftKey,			"DeleteBackward"					},
    {Keyboard_Delete,				0,					"DeleteForward"						},
    {Keyboard_Backspace,			CtrlKey,			"DeleteWordBackward"				},
    {Keyboard_Delete,				CtrlKey,			"DeleteWordForward"					},

	{Keyboard_Left_Arrow,			0,                  "MoveLeft"							},
    {Keyboard_Left_Arrow,			ShiftKey,           "MoveLeftAndModifySelection"		},
    {Keyboard_Left_Arrow,			CtrlKey,            "MoveWordLeft"						},
    {Keyboard_Left_Arrow,			CtrlKey | ShiftKey, "MoveWordLeftAndModifySelection"	},
    {Keyboard_Right_Arrow,			0,                  "MoveRight"							},
    {Keyboard_Right_Arrow,			ShiftKey,           "MoveRightAndModifySelection"		},
    {Keyboard_Right_Arrow,			CtrlKey,            "MoveWordRight"						},
    {Keyboard_Right_Arrow,			CtrlKey | ShiftKey, "MoveWordRightAndModifySelection"	},
    {Keyboard_Up_Arrow,				0,                  "MoveUp"							},
    {Keyboard_Up_Arrow,				ShiftKey,           "MoveUpAndModifySelection"			},
    {Keyboard_Down_Arrow,			0,                  "MoveDown"							},
    {Keyboard_Down_Arrow,			ShiftKey,           "MoveDownAndModifySelection"		},

	{Keyboard_Keypad_Left_Arrow,	0,                  "MoveLeft"							},
    {Keyboard_Keypad_Left_Arrow,	ShiftKey,           "MoveLeftAndModifySelection"		},
    {Keyboard_Keypad_Left_Arrow,	CtrlKey,            "MoveWordLeft"						},
    {Keyboard_Keypad_Left_Arrow,	CtrlKey | ShiftKey, "MoveWordLeftAndModifySelection"	},
    {Keyboard_Keypad_Right_Arrow,	0,                  "MoveRight"							},
    {Keyboard_Keypad_Right_Arrow,	ShiftKey,           "MoveRightAndModifySelection"		},
    {Keyboard_Keypad_Right_Arrow,	CtrlKey,            "MoveWordRight"						},
    {Keyboard_Keypad_Right_Arrow,	CtrlKey | ShiftKey, "MoveWordRightAndModifySelection"	},
    {Keyboard_Keypad_Up_Arrow,		0,                  "MoveUp"							},
    {Keyboard_Keypad_Up_Arrow,		ShiftKey,           "MoveUpAndModifySelection"			},
    {Keyboard_Keypad_Down_Arrow,	0,                  "MoveDown"							},
    {Keyboard_Keypad_Down_Arrow,	ShiftKey,           "MoveDownAndModifySelection"		},
};

#define ARRAYSIZE(array) (sizeof(array) / sizeof((array)[0]))

const char* EditorClientWiiU::interpretKeyEvent(const KeyboardEvent* event)
{
    ASSERT(event->type() == eventNames().keydownEvent || event->type() == eventNames().keypressEvent);

    static HashMap<int, const char*>* keyDownCommandsMap = 0;
    static HashMap<int, const char*>* keyPressCommandsMap = 0;

    if (!keyDownCommandsMap) {
        keyDownCommandsMap = new HashMap<int, const char*>;
        keyPressCommandsMap = new HashMap<int, const char*>;

		ADD_TO_STATIC_LOCAL(keyDownCommandsMap);
		ADD_TO_STATIC_LOCAL(keyPressCommandsMap);

        for (unsigned i = 0; i < ARRAYSIZE(keyDownEntries); i++)
            keyDownCommandsMap->set(keyDownEntries[i].modifiers << 16 | keyDownEntries[i].virtualKey, keyDownEntries[i].name);

        for (unsigned i = 0; i < ARRAYSIZE(keyPressEntries); i++)
            keyPressCommandsMap->set(keyPressEntries[i].modifiers << 16 | keyPressEntries[i].charCode, keyPressEntries[i].name);
    }

    unsigned modifiers = 0;
    if (event->shiftKey())
        modifiers |= ShiftKey;
    if (event->altKey())
        modifiers |= AltKey;
    if (event->ctrlKey())
        modifiers |= CtrlKey;

    if (event->type() == eventNames().keydownEvent) {
        int mapKey = modifiers << 16 | event->keyCode();
        return mapKey ? keyDownCommandsMap->get(mapKey) : 0;
    }

    int mapKey = modifiers << 16 | event->charCode();
    return mapKey ? keyPressCommandsMap->get(mapKey) : 0;
}

bool EditorClientWiiU::handleEditingKeyboardEvent(KeyboardEvent* event)
{
    Node* node = event->target()->toNode();
    Frame* frame = node->document()->frame();

    const PlatformKeyboardEvent* keyEvent = event->keyEvent();
    if (!keyEvent)
        return false;

    if (!frame->settings())
        return false;

    bool caretBrowsing = frame->settings()->caretBrowsingEnabled();
    if (caretBrowsing) {
        switch (keyEvent->windowsVirtualKeyCode()) {
        case VK_LEFT:
            frame->selection()->modify(keyEvent->shiftKey() ? FrameSelection::AlterationExtend : FrameSelection::AlterationMove,
                    DirectionLeft,
                    keyEvent->ctrlKey() ? WordGranularity : CharacterGranularity,
                    UserTriggered);
            return true;
        case VK_RIGHT:
            frame->selection()->modify(keyEvent->shiftKey() ? FrameSelection::AlterationExtend : FrameSelection::AlterationMove,
                    DirectionRight,
                    keyEvent->ctrlKey() ? WordGranularity : CharacterGranularity,
                    UserTriggered);
            return true;
        case VK_UP:
            frame->selection()->modify(keyEvent->shiftKey() ? FrameSelection::AlterationExtend : FrameSelection::AlterationMove,
                    DirectionBackward,
                    keyEvent->ctrlKey() ? ParagraphGranularity : LineGranularity,
                    UserTriggered);
            return true;
        case VK_DOWN:
            frame->selection()->modify(keyEvent->shiftKey() ? FrameSelection::AlterationExtend : FrameSelection::AlterationMove,
                    DirectionForward,
                    keyEvent->ctrlKey() ? ParagraphGranularity : LineGranularity,
                    UserTriggered);
            return true;
        }
    }

    Editor::Command command = frame->editor()->command(interpretKeyEvent(event));

    if (keyEvent->type() == PlatformKeyboardEvent::RawKeyDown) {
        // WebKit doesn't have enough information about mode to decide how commands that just insert text if executed via Editor should be treated,
        // so we leave it upon WebCore to either handle them immediately (e.g. Tab that changes focus) or let a keypress event be generated
        // (e.g. Tab that inserts a Tab character, or Enter).
        return !command.isTextInsertion() && command.execute(event);
    }

    if (command.execute(event))
        return true;

    // Don't insert null or control characters as they can result in unexpected behaviour
    if (event->charCode() < ' ')
        return false;

    // Don't insert anything if a modifier is pressed
    if (keyEvent->ctrlKey() || keyEvent->altKey())
        return false;

    return frame->editor()->insertText(event->keyEvent()->text(), event);
}

void EditorClientWiiU::handleKeyboardEvent(KeyboardEvent* event)
{
    if (handleEditingKeyboardEvent(event))
        event->setDefaultHandled();
}

void EditorClientWiiU::handleInputMethodKeydown(KeyboardEvent* event)
{
}

EditorClientWiiU::EditorClientWiiU(WebView *view)
    : m_view(view)
{
    notImplemented();
}

EditorClientWiiU::~EditorClientWiiU()
{
    notImplemented();
}

void EditorClientWiiU::textFieldDidBeginEditing(Element*)
{
    notImplemented();
}

void EditorClientWiiU::textFieldDidEndEditing(Element*)
{
	notImplemented();
}

void EditorClientWiiU::textDidChangeInTextField(Element*)
{
    notImplemented();
}

bool EditorClientWiiU::doTextFieldCommandFromEvent(Element*, KeyboardEvent*)
{
    return false;
}

void EditorClientWiiU::textWillBeDeletedInTextField(Element*)
{
    notImplemented();
}

void EditorClientWiiU::textDidChangeInTextArea(Element*)
{
    notImplemented();
}

void EditorClientWiiU::ignoreWordInSpellDocument(const String&)
{
    notImplemented();
}

void EditorClientWiiU::learnWord(const String&)
{
    notImplemented();
}

void EditorClientWiiU::checkSpellingOfString(const UChar*, int, int*, int*)
{
    notImplemented();
}

String EditorClientWiiU::getAutoCorrectSuggestionForMisspelledWord(const String&)
{
    notImplemented();
    return String();
}

void EditorClientWiiU::checkGrammarOfString(const UChar*, int, Vector<GrammarDetail>&, int*, int*)
{
    notImplemented();
}

void EditorClientWiiU::updateSpellingUIWithGrammarString(const String&, const GrammarDetail&)
{
    notImplemented();
}

void EditorClientWiiU::updateSpellingUIWithMisspelledWord(const String&)
{
    notImplemented();
}

void EditorClientWiiU::showSpellingUI(bool)
{
    notImplemented();
}

bool EditorClientWiiU::spellingUIIsShowing()
{
    notImplemented();
    return false;
}

void EditorClientWiiU::getGuessesForWord(const String& word, const String& context, Vector<String>& guesses)
{
    notImplemented();
}

}
